@@ -20,9 +20,9 @@
 #ifndef __OPENCONNECT_H__
 #define __OPENCONNECT_H__
 
-#include <stdint.h>
 #include <sys/types.h>
-#include <unistd.h>
+
+#include <stdint.h>
 
 #ifdef __cplusplus
 extern "C" {
@@ -33,9 +33,22 @@ extern "C" {
 #endif
 
 #define OPENCONNECT_API_VERSION_MAJOR 5
-#define OPENCONNECT_API_VERSION_MINOR 5
+#define OPENCONNECT_API_VERSION_MINOR 7
 
 /*
+ * API version 5.7 (v8.20; 2022-02-20):
+ *  - Add openconnect_get_connect_url()
+ *  - Add openconnect_set_cookie()
+ *  - Add openconnect_set_allow_insecure_crypto()
+ *  - Add openconnect_get_auth_expiration()
+ *  - Add openconnect_disable_dtls()
+ *  - Make openconnect_disable_ipv6() return int
+ *  - Add openconnect_set_webview_callback()
+ *  - Add openconnect_webview_load_changed()
+ *
+ * API version 5.6 (v8.06; 2020-03-31):
+ *  - Add openconnect_set_trojan_interval()
+ *
  * API version 5.5 (v8.00; 2019-01-05):
  *  - add openconnect_set_version_string()
  *  - add openconnect_set_key_password()
@@ -182,6 +195,8 @@ extern "C" {
 #define OC_PROTO_AUTH_CERT	(1<<2)
 #define OC_PROTO_AUTH_OTP	(1<<3)
 #define OC_PROTO_AUTH_STOKEN	(1<<4)
+#define OC_PROTO_PERIODIC_TROJAN	(1<<5)
+#define OC_PROTO_HIDDEN	(1<<6)
 
 struct oc_vpn_proto {
 	const char *name;
@@ -272,9 +287,10 @@ struct oc_split_include {
 
 struct oc_ip_info {
 	const char *addr;
-	const char *netmask;
+	const char *netmask; /* Just the netmask, in dotted-quad form. */
 	const char *addr6;
-	const char *netmask6;
+	const char *netmask6; /* This is the IPv6 address *and* netmask
+			       * e.g. "2001::dead:beef/128". */
 	const char *dns[3];
 	const char *nbns[3];
 	const char *domain;
@@ -310,6 +326,17 @@ struct oc_cert {
 	void *reserved;
 };
 
+/* Used by openconnect_webview_load_changed() to return data to OC.
+ * The arrays must contain an even number of const strings, with names and
+ * values, and a NULL to terminate. E.g.:
+ * name0, val0, name1, val1, NULL
+ */
+struct oc_webview_result {
+	const char *uri;
+	const char **cookies;
+	const char **headers;
+};
+
 /****************************************************************************/
 
 #define PRG_ERR		0
@@ -345,6 +372,7 @@ typedef enum {
 	OC_TOKEN_MODE_TOTP,
 	OC_TOKEN_MODE_HOTP,
 	OC_TOKEN_MODE_YUBIOATH,
+	OC_TOKEN_MODE_OIDC,
 } oc_token_mode_t;
 
 typedef enum {
@@ -377,7 +405,7 @@ int openconnect_set_csd_environ(struct openconnect_info *vpninfo,
 /* This string is static, valid only while the connection lasts. If you
  * are going to cache this to remember which certs the user has accepted,
  * make sure you also store the host/port for which it was accepted and
- * don't just accept this cert from *anywhere*. Also use use the check
+ * don't just accept this cert from *anywhere*. Also use the check
  * function below instead of manually comparing. When this function
  * returns a string which *doesn't* match the previously-stored hash
  * matched with openconnect_check_peer_cert_hash(), you should store
@@ -434,11 +462,85 @@ int openconnect_init_ssl(void);
 const char *openconnect_get_cstp_cipher(struct openconnect_info *);
 const char *openconnect_get_dtls_cipher(struct openconnect_info *);
 
-/* These return a descriptive string of the compression algorithm 
+/* These return a descriptive string of the compression algorithm
  * in use (LZS, LZ4, ...). If no compression then NULL is returned. */
 const char *openconnect_get_cstp_compression(struct openconnect_info *);
 const char *openconnect_get_dtls_compression(struct openconnect_info *);
 
+
+/*
+ * Since authentication can run in a separate environment to the connection
+ * itself, there is a simple set of information which needs to be passed
+ * from one to the other. Basically it's just the server to connect to, and
+ * the cookie we need for authentication. And luckily, as we've added more
+ * and more protocols over the years, the "cookie" part has remained true
+ * and we haven't needed to use client certificates for the *connection*.
+ *
+ * The *server* part is a little more complex. Firstly, the certificate
+ * might not be valid and may have been accepted manually by the user,
+ * so we pass the certificate fingerprint separately, as a second piece
+ * of information.
+
+ * In the beginning, we passed the server hostname as the third piece of
+ * information, and all was well.
+ *
+ * Then we found servers on a non-standard port, so the authentication
+ * dialogs would use openconnect_get_port() and just append it (":%d")
+ * to the hostname string they passed on.
+ *
+ * Then we encountered servers with round-robin or geo-DNS, which gave
+ * different IP addresses for a given hostname, and we switched the
+ * openconnect_get_hostname() function to return the *IP* address instead,
+ * since the actual name didn't matter when it wasn't being used to check
+ * the server's certificate anyway.
+ *
+ * At some point later, openconnect_get_dnsname() was added to return the
+ * actual hostname, during the authentication phase where the cert was
+ * being presented to the user for manual acceptance. But that wasn't
+ * really important for the authentication → connection handoff.
+ *
+ * Later still, Legacy IP addresses got scarce and SNI was invented, and
+ * we started to see servers behind proxies that forward a connection
+ * based on the SNI in the incoming ClientHello (TLS layer), or based on
+ * the 'Host' header in the initial connection-phase request (HTTP[S]
+ * layer). So returning just the IP address from openconnect_get_hostname()
+ * now made things break.
+ *
+ * So... we need to pass *both* the actual hostname *and* the IP address
+ * to the connecting openconnect invocation. As well as the port.
+ *
+ * In addition, the Pulse protocol introduced a new requirement for the
+ * connection. Instead of connecting to a fixed endpoint on the server,
+ * we must connect to the appropriate *path*, which varies. So in fact
+ * it isn't just the "hostname" any more, but the full URL.
+ *
+ * So, now we have openconnect_get_connect_url() which gets the full URL
+ * including the port and path, and the original hostname.
+ *
+ * Since we're now back to giving openconnect a hostname, we need to add
+ * a '--resolve' argument to avoid the round robin DNS problem and ensure
+ * that we actually connect to the same server we authenticated to. The
+ * arguments for that can be obtained from openconnect_get_dnsname() and
+ * openconnect_get_hostname() — the latter of which, as noted, was changed
+ * years ago to return a numeric address. We end up invoking openconnect
+ * to make the connection as follows:
+ *
+ * openconnect $CONNECT_URL --servercert $FINGERPRINT --cookie $COOKIE \
+ *             --resolve $DNSNAME:$HOSTNAME
+ *
+ * ... where '$HOSTNAME', as returned by openconnect_get_hostname(),
+ * isn't actually a hostname (as noted above). Sorry.
+ *
+ * In fact, what you get back from openconnect_get_hostname() is the
+ * IP literal in the form it would appear in a URL. So IPv6 addresses
+ * are wrapped in [], and that needs to be *stripped* in order to pass
+ * it to openconnect's --resolve argument. As I realise and type this,
+ * it doesn't seem particularly useful to provide yet another function
+ * that will return the non-[]-wrapped version, as we'd still need UI
+ * tools to do it themselves for backward compatibility. Sorry again :)
+ */
+const char *openconnect_get_connect_url(struct openconnect_info *);
+
 /* Returns the IP address of the exact host to which the connection
  * was made. In --cookieonly mode or in any other scenario involving
  * a "two stage" connection, it is important to reconnect by IP because
@@ -521,7 +623,9 @@ int openconnect_set_key_password(struct openconnect_info *vpninfo, const char *p
 const char *openconnect_get_ifname(struct openconnect_info *);
 void openconnect_set_reqmtu(struct openconnect_info *, int reqmtu);
 void openconnect_set_dpd(struct openconnect_info *, int min_seconds);
+void openconnect_set_trojan_interval(struct openconnect_info *, int seconds);
 int openconnect_get_idle_timeout(struct openconnect_info *);
+time_t openconnect_get_auth_expiration(struct openconnect_info *);
 
 /* The returned structures are owned by the library and may be freed/replaced
    due to rekey or reconnect. Assume that once the mainloop starts, the
@@ -534,14 +638,17 @@ int openconnect_get_ip_info(struct openconnect_info *,
 
 int openconnect_get_port(struct openconnect_info *);
 const char *openconnect_get_cookie(struct openconnect_info *);
+int openconnect_set_cookie(struct openconnect_info *, const char *);
 void openconnect_clear_cookie(struct openconnect_info *);
 
-void openconnect_disable_ipv6(struct openconnect_info *vpninfo);
+int openconnect_disable_ipv6(struct openconnect_info *vpninfo);
+int openconnect_disable_dtls(struct openconnect_info *vpninfo);
 void openconnect_reset_ssl(struct openconnect_info *vpninfo);
 int openconnect_parse_url(struct openconnect_info *vpninfo, const char *url);
 void openconnect_set_cert_expiry_warning(struct openconnect_info *vpninfo,
 					 int seconds);
 void openconnect_set_pfs(struct openconnect_info *vpninfo, unsigned val);
+int openconnect_set_allow_insecure_crypto(struct openconnect_info *vpninfo, unsigned val);
 
 /* If this is set, then openconnect_obtain_cookie() will abort and return
    failure if the file descriptor is readable. Typically a user may create
@@ -638,6 +745,16 @@ struct openconnect_info *openconnect_vpninfo_new(const char *useragent,
 						 void *privdata);
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo);
 
+typedef int (*openconnect_open_webview_vfn) (struct openconnect_info *,
+					     const char *uri,
+					     void *privdata);
+
+void openconnect_set_webview_callback(struct openconnect_info *vpninfo,
+				      openconnect_open_webview_vfn);
+
+int openconnect_webview_load_changed(struct openconnect_info *vpninfo,
+                                     const struct oc_webview_result *result);
+
 /* Callback to allow binding a newly created socket's file descriptor to
    a specific interface, e.g. with SO_BINDTODEVICE. This tells the kernel
    not to route the traffic in question over the VPN tunnel. */
@@ -681,7 +798,7 @@ typedef int (*openconnect_getaddrinfo_vfn) (void *privdata, const char *node, co
 					    const struct addrinfo *hints, struct addrinfo **res);
 void openconnect_override_getaddrinfo(struct openconnect_info *vpninfo, openconnect_getaddrinfo_vfn gai_fn);
 
-/* Callback for configuring the interface after MTU detection finishes. */
+/* Callback for configuring the interface after tunnel is fully up. */
 typedef void (*openconnect_setup_tun_vfn) (void *privdata);
 void openconnect_set_setup_tun_handler(struct openconnect_info *vpninfo,
 				       openconnect_setup_tun_vfn setup_tun);
